
%{
/*************************************************
 *    Lexer.l
 *************************************************/

#undef yywrap
int yywrap();

#include "globals.h"
#include "tok_val.h"
#include "Lexer.h"
#include "symtab.h"

int  yylineno = 1;
int  yycolno = 0;
int  nxt_pos = 0;
char line_buf[MAX_TOKN_LEN];

int  errlineno;
int  errcolno;
char errline[MAX_TOKN_LEN];
int  err_cnt = 0;
int  err_tok = 0;
extern char *cur_file;

int  in_preprocess = 0;
int  in_charconst = 0;

char  tokn_buff[MAX_TOKN_LEN];
char *tokn_end = &tokn_buff[MAX_TOKN_LEN-1];
char *tokn_ptr;

void yywarn  (char *s);
void yyerror (char *s);
void yyerr   (char *s);

#undef YY_INPUT

#    ifdef STRING_PARSE

#define    min(X,Y)    ((X)<(Y) ? (X) : (Y))

    /* Scan input from a string instead */
#define YY_INPUT(buf,result,max_size) \
  do { \
     int n = min(max_size, parseEnv->parselimit - parseEnv->parseptr);\
     if (n>0){ \
        memcpy(buf,parseEnv->parseptr,(unsigned int) n); \
        parseEnv->parseptr += n; \
        result = n; \
     } else \
        result = YY_NULL; \
  } while(0)

#    else    /* not STRING_PARSE */

    /* Scan input from a file pointer (default) */
#define YY_INPUT(buf,result,max_size) \
    do { \
        if (fgets(buf,max_size,infp) == NULL) { \
           result = YY_NULL; \
        } else { \
           result = strlen(buf); \
           strncpy(line_buf, buf, MAX_TOKN_LEN); \
        } \
    } while(0)

#    endif    /* not STRING_PARSE */

#define YY_USER_ACTION \
    { \
        yycolno = nxt_pos; \
        nxt_pos += yyleng; \
    }

#undef PUTS
#define PUTS \
	do { \
          fputs(yytext,outfp); \
	} while(0)


#undef  LAST_STATE
#define LAST_STATE() \
        do { \
          if (in_preprocess) \
            BEGIN(PP); \
          else \
            BEGIN(INITIAL); \
        } while (0)  

#undef  BEGIN_PP
#define BEGIN_PP() \
        do { \
          BEGIN(PP); \
          in_preprocess = 1; \
        } while(0)

YYSTYPE my_var;

EXTERN int yylex();

%}
%x Start PP CMMT CC STR
%a 9000
%o 9000

dot		"."
digit		[0-9]
octdigit	[0-7]
hexdigit	[0-9a-fA-F]
alpha		[a-zA-Z_]
alphanum	{alpha}|{digit}
litquote	\\\"

number		{digit}+
numsuffix       [uU]?[lL]?
intnum		{digit}+{numsuffix}
octnum		0{octdigit}+{numsuffix}
hexnum		0[xX]{hexdigit}+{numsuffix}

frac		{dot}{number}
mantissa	{number}|{number}{dot}|{number}{frac}|{frac}
exponent	[Ee][\+\-]?{number}
realsuffix      [fFlL]
real		{mantissa}{exponent}?{realsuffix}?

whitespace      [ \t\f\v]

pp_strt         ^{whitespace}*"#"{whitespace}*

%%

<INITIAL>"/*"          { BEGIN(CMMT); PUTS;  }
<INITIAL,PP>"'"        { BEGIN(CC);   PUTS; }
<INITIAL,PP>\"         { BEGIN(STR);  tokn_ptr = tokn_buff; PUTS; }

<INITIAL,PP>"//".*$    { /* C++ style comment */
                         PUTS;
                       }

<INITIAL,PP>"auto"     { PUTS; return(AUTO); }
<INITIAL,PP>"break"    { PUTS; return(BREAK); }
<INITIAL,PP>"case"     { PUTS; return(CASE); }
<INITIAL,PP>"char"     { PUTS; return(CHAR); }
<INITIAL,PP>"const"    { PUTS; return(CONST); }
<INITIAL,PP>"continue" { PUTS; return(CONT); }
<INITIAL,PP>"default"  { PUTS; return(DEFLT); }
<INITIAL,PP>"do"       { PUTS; return(DO); }
<INITIAL,PP>"double"   { PUTS; return(DOUBLE); }
<INITIAL,PP>"else"     { PUTS; return(ELSE); }
<INITIAL,PP>"enum"     { PUTS; return(ENUM); }
<INITIAL,PP>"extern"   { PUTS; return(EXTRN); }
<INITIAL,PP>"float"    { PUTS; return(FLOAT); }
<INITIAL,PP>"for"      { PUTS; return(FOR); }
<INITIAL,PP>"goto"     { PUTS; return(GOTO); }
<INITIAL,PP>"if"       { PUTS; return(IF); }
<INITIAL,PP>"int"      { PUTS; return(INT); }
<INITIAL,PP>"long"     { PUTS; return(LONG); }
<INITIAL,PP>"register" { PUTS; return(REGISTR); }
<INITIAL,PP>"return"   { PUTS; return(RETURN); }
<INITIAL,PP>"short"    { PUTS; return(SHORT); }
<INITIAL,PP>"signed"   { PUTS; return(SGNED); }
<INITIAL,PP>"sizeof"   { PUTS; return(SIZEOF); }
<INITIAL,PP>"static"   { PUTS; return(STATIC); }
<INITIAL,PP>"struct"   { PUTS; return(STRUCT); }
<INITIAL,PP>"switch"   { PUTS; return(SWITCH); }
<INITIAL,PP>"typedef"  { PUTS; return(TYPEDEF); }
<INITIAL,PP>"union"    { PUTS; return(UNION); }
<INITIAL,PP>"unsigned" { PUTS; return(UNSGNED); }
<INITIAL,PP>"void"     { PUTS; return(VOID); }
<INITIAL,PP>"volatile" { PUTS; return(VOLATILE); }
<INITIAL,PP>"while"    { PUTS; return(WHILE); }

<INITIAL,PP>"class"     { PUTS; return(1); }
<INITIAL,PP>"public"    { PUTS; return(1); }
<INITIAL,PP>"private"   { PUTS; return(1); }
<INITIAL,PP>"protected" { PUTS; return(1); }
<INITIAL,PP>"new"       { PUTS; return(1); }
<INITIAL,PP>"delete"    { PUTS; return(1); }
<INITIAL,PP>"this"      { PUTS; return(1); }
<INITIAL,PP>"operator"  { PUTS; return(1); }
<INITIAL,PP>"friend"    { PUTS; return(1); }
<INITIAL,PP>"inline"    { PUTS; return(1); }
<INITIAL,PP>"virtual"   { PUTS; return(1); }
<INITIAL,PP>"template"  { PUTS; return(1); }
<INITIAL,PP>"try"       { PUTS; return(1); }
<INITIAL,PP>"throw"     { PUTS; return(1); }
<INITIAL,PP>"catch"     { PUTS; return(1); }

<INITIAL,PP>"+"        { PUTS; return(PLUS); }
<INITIAL,PP>"-"        { PUTS; return(MINUS); }
<INITIAL,PP>"*"        { PUTS; return(STAR); }
<INITIAL,PP>"/"        { PUTS; return(DIV); }
<INITIAL,PP>"%"        { PUTS; return(MOD); }

<INITIAL,PP>"+="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"-="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"*="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"/="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"%="       { PUTS; return(ASSIGN); }

<INITIAL,PP>"!"        { PUTS; return(NOT); }
<INITIAL,PP>"&&"       { PUTS; return(AND); }
<INITIAL,PP>"||"       { PUTS; return(OR); }
<INITIAL,PP>"^^"       { PUTS; return(XOR); }

<INITIAL,PP>"&&="      { PUTS; return(ASSIGN); }
<INITIAL,PP>"||="      { PUTS; return(ASSIGN); }
<INITIAL,PP>"^^="      { PUTS; return(ASSIGN); }

<INITIAL,PP>"~"        { PUTS; return(B_NOT); }
<INITIAL,PP>"&"        { PUTS; return(B_AND); }
<INITIAL,PP>"|"        { PUTS; return(B_OR); }
<INITIAL,PP>"^"        { PUTS; return(B_XOR); }

<INITIAL,PP>"~="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"&="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"|="       { PUTS; return(ASSIGN); }
<INITIAL,PP>"^="       { PUTS; return(ASSIGN); }

<INITIAL,PP>"<<"       { PUTS; return(L_SHIFT); }
<INITIAL,PP>">>"       { PUTS; return(R_SHIFT); }
<INITIAL,PP>"<<="      { PUTS; return(ASSIGN); }
<INITIAL,PP>">>="      { PUTS; return(ASSIGN); }

<INITIAL,PP>"=="       { PUTS; return(COMP_EQ); }
<INITIAL>"<"           { PUTS; return(COMP_ARITH); }
<PP>"<"                { PUTS; return(LESS); }
<INITIAL,PP>"<="       { PUTS; return(COMP_ARITH); }
<INITIAL>">"           { PUTS; return(COMP_ARITH); }
<PP>">"                { PUTS; return(GRTR); }
<INITIAL,PP>">="       { PUTS; return(COMP_ARITH); }
<INITIAL,PP>"!="       { PUTS; return(COMP_EQ); }

<INITIAL,PP>"="        { PUTS; return(EQ); }
<INITIAL,PP>"++"       { PUTS; return(INCR); }
<INITIAL,PP>"--"       { PUTS; return(DECR); }

<INITIAL,PP>"("        { PUTS; return(LPAREN); }
<INITIAL,PP>")"        { PUTS; return(RPAREN); }
<INITIAL,PP>"["        { PUTS; return(LBRCKT); }
<INITIAL,PP>"]"        { PUTS; return(RBRCKT); }
<INITIAL,PP>"{"        { PUTS; return(LBRACE); }
<INITIAL,PP>"}"        { PUTS; return(RBRACE); }

<INITIAL,PP>"."        { PUTS; return(DOT); }
<INITIAL,PP>"->"       { PUTS; return(ARROW); }

<INITIAL,PP>"?"        { PUTS; return(QUESTMARK); }
<INITIAL,PP>":"        { PUTS; return(COLON); }
<INITIAL,PP>";"        { PUTS; return(SEMICOLON); }
<INITIAL,PP>","        { PUTS; return(COMMA); }
<INITIAL,PP>"..."      { PUTS; return(ELIPSIS); }

	/* Unused (invalid) characters */
<INITIAL,PP>"`"        { PUTS; return(BACKQUOTE); }
<INITIAL,PP>"@"        { PUTS; return(AT); }
<INITIAL,PP>"$"        { PUTS; return(DOLLAR); }

<INITIAL,PP>"#"        { PUTS; return(LB_SIGN); }
<INITIAL,PP>"##"       { PUTS; return(DOUB_LB_SIGN); }

	/* Preprocessor Stuff */
<INITIAL>{pp_strt}\n       { yylineno++; PUTS; nxt_pos = 0; }
<INITIAL>{pp_strt}"define" { BEGIN_PP(); PUTS; return(PP_DEFINE); }
<PP>"defined"              { PUTS; return(PP_DEFINED); }
<INITIAL>{pp_strt}"elif"   { BEGIN_PP(); PUTS; return(PP_ELIF); }
<INITIAL>{pp_strt}"else"   { BEGIN_PP(); PUTS; return(PP_ELSE); }
<INITIAL>{pp_strt}"endif"  { BEGIN_PP(); PUTS; return(PP_ENDIF); }
<INITIAL>{pp_strt}"error"  { BEGIN_PP(); PUTS; return(PP_ERROR); }
<INITIAL>{pp_strt}"if"     { BEGIN_PP(); PUTS; return(PP_IF); }
<INITIAL>{pp_strt}"ifdef"  { BEGIN_PP(); PUTS; return(PP_IFDEF); }
<INITIAL>{pp_strt}"ifndef" { BEGIN_PP(); PUTS; return(PP_IFNDEF); }
<INITIAL>{pp_strt}"include" { BEGIN_PP(); PUTS; return(PP_INCL); }
<INITIAL>{pp_strt}"line"   { BEGIN_PP(); PUTS; return(PP_LINE); }
<INITIAL>{pp_strt}"pragma" { BEGIN_PP(); PUTS; return(PP_PRAGMA); }
<INITIAL>{pp_strt}"undef"  { BEGIN_PP(); PUTS; return(PP_UNDEF); }

<PP>\\(\n)             {  /* Preprocessor continuation line */
                         PUTS;
                         yylineno++;
                         nxt_pos = 0;
                       }

<PP>\n                 {  /* End of this preprocessor logical line */
                         in_preprocess = 0;
                         yyless(0);    /* Hack! */
                         BEGIN(INITIAL);
                         return(PP_END);
                       }

<INITIAL,PP>"__LINE__" { PUTS; return(LINE_NUM); }
<INITIAL,PP>"__FILE__" { PUTS; return(FILE_NME); }
<INITIAL,PP>"__DATE__" { PUTS; return(DATE); }
<INITIAL,PP>"__TIME__" { PUTS; return(TIME); }
<INITIAL,PP>"__STDC__" { PUTS; return(STDC); }

<INITIAL,PP>[a-zA-Z_][a-zA-Z0-9_]*  {    /* Identifier */
                         fputs(symlook(yytext,yyleng+1),outfp);
#ifdef    ENTER_SYM
                         syminsert(yytext,yytext);
#endif
                         return(IDENT);
                       }

<INITIAL,PP>{octnum}  |
<INITIAL,PP>{intnum}  |
<INITIAL,PP>{hexnum}   {  /* An integer */
                         PUTS;
                         return(INUM);
                       }
          
<INITIAL,PP>[0-9]*"."?[0-9]*[eE][+-]?[0-9]+? |
<INITIAL,PP>[0-9]*"."[0-9]+([eE][+-]?[0-9]+)? {
                         PUTS;
                         return(RNUM);
                       }

<STR>\"                { /* Closing quote */
                         PUTS;
                         LAST_STATE();
                         return(STRING);
                       }

<STR>\n                { /* Error - unterminated string constant */
                         PUTS;
                         yyerr("Unterminated string constant starting");
                         LAST_STATE();
                         return(INVALID);
                       }

<STR>\\[0-7]{1,3}      {   /* octal escape sequence */
                         PUTS;
                         return(INVALID);
                       }

<STR>\\x{hexdigit}+    {   /* hex escape sequence - ISO C */
                         PUTS;
                         return(INVALID);
                       }

<STR>\\{number}        {  /* Bad escape sequence */
                         PUTS;
                         return(INVALID);
                       }

<STR>\\n               {
                         PUTS;
                         return(1);
                       }

<STR>\\t               {
                         PUTS;
                         return(1);
                       }

<STR>\\r               {
                         PUTS;
                         return(1);
                       }

<STR>\\b               {
                         PUTS;
                         return(1);
                       }

<STR>\\f               {
                         PUTS;
                         return(1);
                       }

<STR>\\v               {
                         PUTS;
                         return(1);
                       }

<STR>\\a               {
                         PUTS;
                         return(1);
                       }

<STR>\\(\n)            {  /* String continuation */
                         PUTS;
                         yylineno++;
                         nxt_pos = 0;
                         return(1);
                       }

<STR>\\.               {
                         PUTS;
                         return(1);
                       }

<STR>[^\\\n\"]+        {
                         PUTS;
                         return(1);
                       }

<CMMT>[^*\n/]*         {   /* Inside C-style comment */
                        PUTS;
                       }
<CMMT>[^*\n/]*\n       { PUTS; yylineno++; nxt_pos = 0; }
<CMMT>"/"[^*\n]        { PUTS; }
<CMMT>"/"\n            { PUTS; yylineno++; nxt_pos = 0; }
<CMMT>"/*"             { PUTS; yywarn("/* inside comment"); }
<CMMT>"*"+[^*/\n]*     {  /* Stars */
                        PUTS;
                       }
<CMMT>"*"+[^*/\n]*\n   { PUTS; yylineno++; nxt_pos = 0; }
<CMMT>"*"+"/"          {
                        PUTS;
                        LAST_STATE();
                       }



<CC>\\[0-7]{1,3}"'"    {   /* octal escape sequence */
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\x{hexdigit}+"'"  {   /* hex escape sequence - ISO C */
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\{number}"'"      {  /* Bad escape sequence */
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\n                 { /* Error unterminated char constant */
                         PUTS;
                         yylineno++;
                         nxt_pos = 0;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\n"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\t"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\r"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\b"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\f"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\v"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\a"'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\\n"'"            {
                         PUTS;
                         yylineno++;
                         nxt_pos = 1;  /* Oddity */
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>\\."'"             {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>."'"               {
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>"'"                { /* Empty */
                         PUTS;
                         yywarn("Empty character constant");
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>[^\\\n].+"'"        { /* Multiple characters */
                         PUTS;
                         LAST_STATE();
                         return(CHAR_CONST);
                       }

<CC>.                  {
                        PUTS;
                       }

<INITIAL,PP>{whitespace}+ { PUTS;/* space/tab/formfeed/vertical tab */}

<INITIAL,CMMT>\n|\r    { PUTS; yylineno++; nxt_pos = 0; }

<INITIAL,PP><<EOF>>    { 
                         PUTS;
                         in_preprocess = 0;
                         return(0);
                       }

<CMMT><<EOF>>          { 
                         PUTS;
                         yyerr("EOF reached inside comment");
                         return(0);
                       }

<CC><<EOF>>            { 
                         PUTS;
                         yyerr("EOF reached inside character constant");
                         return(0);
                       }

<STR><<EOF>>           { 
                         PUTS;
                         yyerr("EOF reached inside string constant");
                         return(0);
                       }

.                      { /* Any unknown char is an error */
                         PUTS;
                        fprintf(stderr,
                            "Error: Line %d: Illegal Character", yylineno);
                        if ((yytext[0] >= ' ') && (yytext[0] <= '~'))
                            fprintf(stderr,": '%c'\n", yytext[0]);
                        else
                            fprintf(stderr,
                                ", ASCII: %03o (octal)\n", yytext[0]);
                       }

%%

/************************************************************************
//
// yywrap() - This function is called by the lexer [yylex() or sslex()] when
//            the end-of-file (or end-of-string for sslex()) is reached.
//            It gets the next file/string ready (if any) and returns 1
//            which indicates to the lexer that there are no more files
//            remaining.  The lexer returns 0 to the parser, indicating
//            no more tokens are remaining. The parser function, yyparse(),
//            then returns 0, indicating that the parse is done.  This
//            behaviour allows a single item to be parsed to be placed
//            on the stack and parsed without the entire stack being consumed.
//
// **********************************************************************/

int yywrap()
{
    in_preprocess = 0;
    return(1);
}

/***********************************************************************/

void print_caret(char *s, int pos)
{
    char *str;
    int i = 0;

    for (str=s; *str && (i<pos); i++, str++)
        if (*str=='\t')
            fputs("\t",stderr);
        else
            fputs(" ",stderr);

    fputs("^\n",stderr);
}
 
/***********************************************************************/

void yywarn(char *s)
{
        /* Insure temination */
    line_buf[MAX_TOKN_LEN-1] = '\0';
    fprintf(stderr,"Warning: %s at line %d:\n%s", s, errlineno, line_buf );
    print_caret(line_buf,yycolno);
}

/***********************************************************************/

void yyerr(char *s)
{
    fprintf(stderr,"%s:%d: %s:\n%s",
         cur_file, yylineno, s, line_buf); 
    print_caret(line_buf,yycolno);

    if (++err_cnt >= 10){
        fprintf(stderr,"Too many errors (%d detected) - quitting.\n",
                err_cnt );
        exit(1);
    }
}

/***********************************************************************/
