
/*  ###############################################################
    ##
    ##     (Gnu?) C Interpreter
    ##
    ##     File:         symtab.c
    ##
    ##     Programmer:   Shawn Flisakowski
    ##     Date:         Jan 11, 1995
    ##
    ##
    ###############################################################  */

#include    <stdio.h>
#include    <string.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "globals.h"
#include    "Lexer.h"
#include    "macro.h"
#include    "symtab.h"

HashItem  *SymTab[MAX_HASH_BCKTS];

/*  ###############################################################  */

int  symhash( str_t *sym )
{
    if (sym)
        return(sym->hash);
    else
        return(-1);
}

/*  ###############################################################  */

char *symstr( str_t *sym )
{
    if (sym)
        return(sym->str);
    else
        return(NULL);
}

/*  ###############################################################  */

int  calc_hash( char *str )
{
int   hsh=0;
char *ch;

    for (; *str; str++){
        hsh += *str;
        hsh << 1;
    }

    return(hsh);
}

/*  ###############################################################  */

void init_symtab()
{
int j;

    for (j=0; j<MAX_HASH_BCKTS; j++)
        SymTab[j] = (HashItem *) NULL;
}

/*  ###############################################################  */

void  free_symtab()
{
int j;
HashItem  *hptr, *hfree;

    for (j=0; j<MAX_HASH_BCKTS; j++){

        hfree = (HashItem *) NULL;

        for (hptr=SymTab[j]; hptr; hptr=hptr->next){
            if (hptr->sym.str)
                free(hptr->sym.str);
            if (hfree)
                free(hfree);
            hfree = hptr;
        }

        if (hfree)
            free(hfree);
    }
}

/*  ###############################################################  */

int syminsert( char *sym, char *replce )
{
int  hsh, bckt;
HashItem *hptr;

    hsh  = calc_hash(sym);
    bckt = hsh % MAX_HASH_BCKTS;

    for (hptr=SymTab[bckt]; hptr; hptr=hptr->next){

        if ((hptr->sym.hash == hsh) && (STRCMP(sym,hptr->sym.str) == 0)){
#ifndef    ENTER_SYM
            fprintf(stderr,"Error: Symbol %s already in table (as %s).\n",
                sym, hptr->replce.str);
            exit(0);
#endif
            return(0);
        }
    } 

    if ((hptr = (HashItem *) malloc( HASH_ITEM_SZE )) == NULL){
        fprintf(stderr,"Error: Out of memory (attempting to insert: %s)\n",
                sym);
        exit(0);
    }

    hptr->sym.hash = hsh;
    hptr->sym.str = strdup(sym);

    hptr->replce.hash = hsh;
    hptr->replce.str = strdup(replce);

    hptr->next = SymTab[bckt];
    SymTab[bckt] = hptr;
    return(1);
}

/*  ###############################################################  */
char *symlook( char *sym, int len )
{
int  hsh, bckt;
HashItem *hptr;

    hsh  = calc_hash(sym);
    bckt = hsh % MAX_HASH_BCKTS;

    for (hptr=SymTab[bckt]; hptr; hptr=hptr->next){

        if ((hptr->sym.hash == hsh) && (STRCMP(sym,hptr->sym.str) == 0))
            return(hptr->replce.str);
    } 

    return(sym);
}

/*  ###############################################################  */

void symshow()
{
int  j;
HashItem *hptr;

    fprintf(stdout,"        Symbol Table\n");
    fprintf(stdout,"       --------------\n\n");

    for (j=0; j<MAX_HASH_BCKTS; j++){
        for (hptr=SymTab[j]; hptr; hptr=hptr->next){

            fprintf(stdout, "Replace '%s' with '%s'\n",
                hptr->sym.str, hptr->replce.str);
        } 
    } 

    fputs("\n\n",stdout);
}

void symshow2()
{
int  j;
HashItem *hptr;

    fprintf(stdout,"        Symbol Table\n");
    fprintf(stdout,"       --------------\n\n");

    for (j=0; j<MAX_HASH_BCKTS; j++){
        for (hptr=SymTab[j]; hptr; hptr=hptr->next){

            fprintf(stdout, "%-38s  cr%s\n",
                hptr->sym.str, hptr->replce.str);
        } 
    } 

    fputs("\n\n",stdout);
}

/*  ###############################################################  */
