
/*  ###############################################################
    ##
    ##     File:         transform.c
    ##
    ##     Programmer:   Shawn Flisakowski
    ##     Date:         Dec 16, 1994
    ##
    ##
    ###############################################################  */

#include    <stdio.h>
#include    <string.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "transform.h"
#include    "globals.h"
#include    "Lexer.h"
#include    "tok_val.h"

#include    "symtab.h"

/*  ###############################################################  */

char  *cur_file;
extern int yylex();

/*  ###############################################################  */

void Usage(char *prog)
{

    fprintf(stderr,"Usage: %s [options] [filename(s)]\n\n");

    fprintf(stderr,
    "This program replaces identifiers and types in C and C++ code.\n\n");

    fprintf(stderr,"Options:\n");
    fprintf(stderr,"\tGeneral:\n");
    fprintf(stderr,"\t-help: \t Show this message (-h).\n");
    
    fprintf(stderr,"\n\tInput/Output:\n");
    fprintf(stderr,"\t-t:     Next arg is the name change file.\n");

    exit(0);
}

/*  ###############################################################  */

int processArgs( int argc, char **argv )
{
    int num_files = 0;
    int value;
    char *pname;
    char *arg;

    pname = *argv++;
    argc--; 

    while(argc-- > 0){

        arg = *argv;
        value = ON;

        if (*arg == '-'){
            arg++;

            if (  (strcmp(arg,"help") == 0) ||
                  (strcmp(arg,"h") == 0) ||
                  (strcmp(arg,"H") == 0) ||
                  (strcmp(arg,"?") == 0)){
                Usage(pname);
            } else
            
            if (strcmp(arg,"t") == 0) {
                argv++; argc--;
                namef = *argv;
            } else

            {    /* Unrecognized option - ignore */
                fprintf(stderr,"Option: '%s' not recognized.\n",arg); 
            }

        } else {
            file_list[num_files++] = arg;    
        }
        argv++;
    }

    file_list[num_files] = NULL;

    return(num_files);
}

/*  ###############################################################  */

void read_names()
{
    char buf[201];
    char *str, *replce;
    char *sep = " \t\n";
    int err = 0;

    init_symtab();

    while(fgets(buf,200,nmefp) != NULL){

        if (*buf == '#')
            continue;

        if (! (str = strtok(buf,sep)))
            continue;

        if (! (replce = strtok(NULL,sep))){
            fprintf(stderr,"No replacement string specified for string: %s\n",
                str);
            err++;
            continue;
        }

        syminsert(str,replce);
    }

    if (err){
        fprintf(stderr,"Missing %d replacement strings - exiting.\n",err);
        exit(1);
    }

    /*  symshow();  */
}

/*  ###############################################################  */

int main( int argc, char **argv )
{
int   j;
char buff[100], cmmd[200];


    if ((processArgs(argc,argv)) < 0)
    	return(-1);

    if (namef){
        if  ((nmefp = fopen(namef,"r")) == NULL){
                fprintf(stderr,
                    "Could not open file '%s' as name switch file.\n",
                        namef);
            exit(1);
        }
    } else if ((nmefp = fopen("name.chnge","r")) == NULL){
        fprintf(stderr,
         "Could not open default file 'name.chnge' as name switch file.\n");
        exit(1);
    }

        /* Read in name switch file */
    read_names();
    fclose(nmefp);

    fprintf(stdout,"Transforming:\n");

    for (j=0, cur_file =file_list[0]; cur_file; j++, cur_file=file_list[j]){

        fprintf(stdout,"%s%s", cur_file, ((j%4 == 3) ? "\n" : " "));

        sprintf(buff,"%s.bak", cur_file); 
        sprintf(cmmd,"/bin/cp %s %s", cur_file, buff);
        system(cmmd);
   
        if ((infp = fopen(buff,"r")) == NULL){
            fprintf(stderr,"Could not open %s for reading.\n",buff);
            continue;
        }

        if ((outfp = fopen(cur_file,"w")) == NULL){
            fprintf(stderr,"Could not open %s for writing.\n",cur_file);
            continue;
        }
       
        while(yylex())
            ;

        fclose(outfp);
        fclose(infp);
    }

    fputs("\n",stdout);
    free_symtab();
    return(0);
}

/*  ###############################################################  */
